package com.infiauto.datastr.auto;

import java.io.Serializable;
import java.util.*;

/**
 *
 * @author court
 */
public class LevenshteinAutomaton
        extends Automaton<String>
        implements Serializable
{
    private enum Parameter { I, M; } // enum Parameter
    private enum Type { USUAL, T, MS; } // enum Type
    private class Position
            implements Comparable<Position>, Serializable
    {
        private Parameter parameter;
        private Type type;
        private int index;
        private int error;
        public Position(Parameter parameter,
                Type type,
                int index,
                int error)
        {
            this.parameter = parameter;
            this.type = type;
            this.index = index;
            this.error = error;
        } // Position
        public Parameter getParameter() { return parameter; } // getParameter
        public Type getType() { return type; } // getType
        public int getIndex() { return index; } // getIndex
        public int getError() { return error; } // getError
        public boolean equals(Object o)
        {
            if(o instanceof Position) return compareTo((Position) o) == 0;
            return false;
        } // equals
        public int compareTo(Position p)
        {
            if((p.getParameter() == parameter)
                    && (p.getType() == type)
                    && (p.getIndex() == index)
                    && (p.getError() == error))
                return 0;
            if(parameter != p.getParameter())
                return (parameter == Parameter.I ? -1 : 1);
            if(type != p.getType())
                return (type == Type.USUAL ? -1 : 1);
            if(index < p.getIndex()) return -1;
            else if(index > p.getIndex()) return 1;
            if(error < p.getError()) return -1;
            else if(error > p.getError()) return 1;
            return 0;
        } // compareTo
        public int hashCode()
        {
            return getParameter().hashCode()
                    ^ getType().hashCode()
                    ^ (getIndex() * 0x00010000)
                    ^ getError();
        } // hashCode
    } // class Position
    private class PositionState
        extends LinkedList<Position>
        implements Comparable<PositionState>, Serializable
    {
        public PositionState() { super(); } // PositionState()
        public PositionState(Collection<? extends Position> c)
        {
            super(c);
        } // PositionState(Collection<? extends Position>)
        public boolean contains(Position p)
        {
            for(Position t : this)
                if(t.equals(p))
                    return true;
            return false;
        } // contains
        public boolean equals(Object o)
        {
            if(o instanceof PositionState) return equals((PositionState) o);
            return false;
        } // equals(Object)
        public boolean equals(PositionState s)
        {
            if(s.size() != size()) return false;
            for(Position p : s)
                if(!this.contains(p))
                    return false;
            return true;
        } // equals(PositionState)
        public int compareTo(PositionState s)
        {
            if(size() < s.size()) return -1;
            else if(size() > s.size()) return 1;
            if(equals(s)) return 0;
            return -1;
        } // compareTo
        public int hashCode()
        {
            int hash = 0;
            for(Position p : this)
                hash ^= p.hashCode();
            return hash;
        } // hashCode
    } // class PositionState
    private class Point
    {
        private Type type;
        private int x;
        private int y;
        public Point(Type type,
                int x,
                int y)
        {
            this.type = type;
            this.x = x;
            this.y = y;
        } // Point
        public Type getType() { return type; } // getType
        public int getX() { return x; } // getX
        public int getY() { return y; } // getY
        public boolean equals(Point p)
        {
            return (p.getType() == type)
                    && (p.getX() == x)
                    && (p.getY() == y);
        } // equals
    } // class Point
    private class SetOfPoints extends HashSet<Point>
    {
        SetOfPoints() { super(); } // SetOfPoints()
        SetOfPoints(Collection<? extends Point> c)
        {
            super(c);
        } // SetOfPoints(Collection<? extends Point>)
    } // class SetOfPoints
    private class Transition
        implements Serializable
    {
        private PositionState from;
        private boolean[] b;
        private PositionState to;
        public Transition(PositionState from,
                boolean[] b,
                PositionState to)
        {
            this.from = from;
            this.b = b;
            this.to = to;
        } // Transition
        public PositionState getFrom() { return from; } // getFrom
        public boolean[] getString() { return b; } // getString
        public PositionState getTo() { return to; } // getTo
    } // class Transition
    private enum ChiType { EPSILON, T, MS; } // enum ChiType

    private static final char NONALPHABET_CHARACTER = '$';
    private final PositionState START_STATE = new PositionState(Arrays.asList(new Position[] { new Position(Parameter.I, Type.USUAL, 0, 0) }));
    private static final ChiType DEFAULT_CHI = ChiType.EPSILON;

    private Map<PositionState,State<String>> state_mappings;
    private DictionaryAutomaton dictionary_automaton;
    private int edit_distance;
    private List<Transition> transitions;
    private ChiType chi;
    private Map<State<String>,Set<Integer>> index_mapping;

    private List<boolean[]> buildPowerSet(int len)
    {
        int max = new Double(Math.pow(2.0, (double) len)).intValue();
        List<boolean[]> result = new ArrayList<boolean[]>(max);
        boolean[] temp = new boolean[len];
        for(int j = 0; j < temp.length; j++) temp[j] = false;
        result.add(temp);

        for(int i = 0; i < max - 1; i++)
        {
            temp = new boolean[len];
            for(int j = 0; j < len; j++) temp[j] = result.get(i)[j];
            int j = len - 1;
            while(j >= 0)
            {
                temp[j] ^= true;
                if(temp[j] == false)
                    j--;
                else
                    break;
            }
            result.add(temp);
        }

        if(len > 1) result.addAll(buildPowerSet(len - 1));
        
        return result;
    } // buildPowerSet

    private Position functionRM(PositionState state)
    {
        Position rm = null;

        for(Position pi : state)
        {
            if(pi.getType() == Type.USUAL)
            {
                rm = pi;
            }
        }

        for(Position pi : state)
        {
            if((pi.getType() == Type.USUAL) && ((pi.getIndex() - pi.getError()) > (rm.getIndex() - rm.getError())))
            {
                rm = pi;
            }
        }

        return rm;
    } // functionRM

    private PositionState functionM(int edit_distance, PositionState state, int string_length)
    {
        PositionState m = new PositionState();

        for(Position pi : state)
        {
            if(pi.getParameter() == Parameter.I)
            {
                m.add(new Position(Parameter.M, pi.getType(), pi.getIndex() + edit_distance + 1 - string_length, pi.getError()));
            }
            else
            {
                m.add(new Position(Parameter.I, pi.getType(), pi.getIndex() - edit_distance - 1 + string_length, pi.getError()));
            }
        }

        return m;
    } // functionM

    private boolean functionF(int edit_distance, Position pos, int string_length)
    {
        if(pos == null) return false;
        if(pos.getParameter() == Parameter.I)
        {
            return ((string_length <= (2 * edit_distance + 1)) && (pos.getError() <= (pos.getIndex() + 2 * edit_distance + 1 - string_length)));
        }
        return (pos.getError() > (pos.getIndex() + edit_distance));
    } // functionF
    
    private SetOfPoints deltaED(int edit_distance, Point point, boolean[] b)
    {
        int index = point.getX();
        int error = point.getY();

        switch(chi)
        {
        case EPSILON:

            // zero-length binary string
            if(b.length == 0)
            {
                if(error < edit_distance)
                {
                    return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1) }));
                }
                return new SetOfPoints();
            }

            // first bit in the binary string is true
            if(b[0] == true)
            {
                return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index + 1, error) }));
            }

            // binary string of length one
            if(b.length == 1)
            {
                if(error < edit_distance)
                {
                    return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                        new Point(Type.USUAL, index + 1, error + 1) }));
                }
                return new SetOfPoints();
            }

            // find the first true bit in the binary string
            int first_one = 0;
            for(int i = 1; i < b.length; i++)
            {
                if(b[i] == true)
                {
                    first_one = i;
                    break;
                }
            }

            if(first_one == 0)
            {
                // no true bit was found
                return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                    new Point(Type.USUAL, index + 1, error + 1) }));
            }

            // found at least one true bit
            first_one++;
            return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                new Point(Type.USUAL, index + 1, error + 1),
                new Point(Type.USUAL, index + first_one, error + first_one - 1) }));

        case T:
            
            if(point.getType() == Type.T)
            {
                if(b[0] == true)
                {
                    return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index + 2, error) }));
                }
                return new SetOfPoints();
            }

            if(b.length == 0)
            {
                if(error < edit_distance)
                {
                    return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1) }));
                }
                return new SetOfPoints();
            }

            if(b[0] == true)
            {
                return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index + 1, error) }));
            }

            if(b.length == 1)
            {
                if(error < edit_distance)
                {
                    return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                        new Point(Type.USUAL, index + 1, error + 1) }));
                }
                return new SetOfPoints();
            }
            
            if(b[1] == true)
            {
                return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                    new Point(Type.USUAL, index + 1, error + 1),
                    new Point(Type.USUAL, index + 2, error + 1),
                    new Point(Type.T, index, error + 1) }));
            }

            // find the first true bit in the binary string
            first_one = 0;
            for(int i = 1; i < b.length; i++)
            {
                if(b[i] == true)
                {
                    first_one = i;
                    break;
                }
            }

            if(first_one == 0)
            {
                return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                    new Point(Type.USUAL, index + 1, error + 1) }));
            }

            return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                new Point(Type.USUAL, index + 1, error + 1),
                new Point(Type.USUAL, index + first_one, error + first_one - 1)}));

        case MS:

            if(point.getType() == Type.MS)
            {
                return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index + 1, error) }));
            }

            if(b.length == 0)
            {
                if(error < edit_distance)
                {
                    return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1) }));
                }
                return new SetOfPoints();
            }

            if(b.length == 1)
            {
                if(error < edit_distance)
                {
                    return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                        new Point(Type.USUAL, index + 1, error + 1),
                        new Point(Type.MS, index, error + 1) }));
                }
            }

            return new SetOfPoints(Arrays.asList(new Point[] { new Point(Type.USUAL, index, error + 1),
                new Point(Type.USUAL, index + 1, error + 1),
                new Point(Type.USUAL, index + 2, error + 1),
                new Point(Type.MS, index, error + 1) }));
        }

        return null;
    } // deltaED

    private boolean[] functionR(int edit_distance, Position pos, boolean[] b)
    {
        int length = -1;

        if(pos.getParameter() == Parameter.I)
        {
            if((edit_distance - pos.getError() + 1) < (b.length - edit_distance - pos.getIndex()))
            {
                length = edit_distance - pos.getError() + 1;
            }
            else
            {
                length = b.length - edit_distance - pos.getIndex();
            }

            boolean[] result = new boolean[length];
            int start = edit_distance + pos.getIndex();// + 1;
            for(int i = 0; i < length; i++)
                result[i] = b[start + i];
            return result;
        }

        if((edit_distance - pos.getError() + 1) < -pos.getIndex())
        {
            length = edit_distance - pos.getError() + 1;
        }
        else
        {
            length = -pos.getIndex();
        }

        boolean[] result = new boolean[length];
        int start = b.length + pos.getIndex();// + 1;
        for(int i = 0; i < length; i++)
            result[i] = b[start + i];
        return result;
    } // functionR

    private PositionState deltaE(int edit_distance, Position q, boolean[] b)
    {
        boolean[] h = functionR(edit_distance, q, b);
        SetOfPoints delta_ed = deltaED(edit_distance, new Point(q.getType(), q.getIndex(), q.getError()), h);
        
        if(delta_ed.isEmpty())
        {
            return new PositionState();
        }
        
        PositionState state = new PositionState();
        
        if(q.getParameter() == Parameter.I)
        {
            for(Point pi : delta_ed)
            {
                state.add(new Position(Parameter.I, pi.getType(), pi.getX() - 1, pi.getY()));
            }
        }
        else
        {
            for(Point pi : delta_ed)
            {
                state.add(new Position(Parameter.M, pi.getType(), pi.getX(), pi.getY()));
            }
        }

        return state;
    } // deltaE

    private PositionState delta(int edit_distance, PositionState state, boolean[] b)
    {
        PositionState next_state = new PositionState();
        boolean add = false;

        for(Position q : state)
        {
            PositionState delta_e = deltaE(edit_distance, q, b);

            if(!delta_e.isEmpty())
            {
                for(Position pi : delta_e)
                {
                    // initialize the add flag to true
                    add = true;

                    Iterator<Position> iter = next_state.iterator();
                    while(iter.hasNext())
                    {
                        Position p = iter.next();

                        if(lessThanSubsume(pi, p))
                        {
                            iter.remove();
                        }
                        else
                        {
                            if(p.equals(pi) || lessThanSubsume(p, pi))
                            {
                                add = false;
                                break;
                            }
                        }
                    }

                    if(add == true)
                    {
                        next_state.add(pi);
                    }
                }
            }
        }

        if(functionF(edit_distance, functionRM(next_state), b.length))
        {
            next_state = functionM(edit_distance, next_state, b.length);
        }

        return next_state;
    } // delta

    private boolean contains(Collection<PositionState> c,
            PositionState s)
    {
        for(PositionState t : c)
            if(s.equals(t))
                return true;
        return false;
    } // contains

    private boolean lessThanSubsume(Position q1, Position q2)
    {
        if((q1.getType() != Type.USUAL) || (q2.getError() <= q1.getError()))
        {
            return false;
        }

        int m = -1;
        if(q2.getType() == Type.T)
        {
            m = q2.getIndex() + 1 - q1.getIndex();
        }
        else
        {
            m = q2.getIndex() - q1.getIndex();
        }
        if(m < 0) m = -m;

        return (m <= (q2.getError() - q1.getError()));
    } // lessThanSubsume

    private boolean coversAllPositions(int edit_distance, int string_length, PositionState state)
    {
        Position pos = state.getFirst();

        if(pos.getParameter() == Parameter.I)
        {
            if(state.equals(START_STATE))
            {
                return (string_length >= (pos.getIndex() + edit_distance));
            }
            else
            {
                for(Position pi : state)
                {
                    if(string_length < (2 * edit_distance + pi.getIndex() - pi.getError() + 1))
                    {
                        return false;
                    }
                }
            }
        }
        else
        {
            Position q = null;
            if(string_length < edit_distance)
            {
                q = new Position(Parameter.M, Type.USUAL, 0, edit_distance - string_length);
            }
            else
            {
                q = new Position(Parameter.M, Type.USUAL, edit_distance - string_length, 0);
            }

            for(Position pi : state)
            {
                if(!pi.equals(q) && !lessThanSubsume(q, pi))
                {
                    return false;
                }
            }
        }

        return true;
    } // coversAllPositions

    private boolean[] generateCharacteristicVector(char c, String s)
    {
        boolean[] characteristic_vector = new boolean[((2 * edit_distance + 2) > s.length()) ? s.length() : (2 * edit_distance + 2)];
        int i = 0;
        for(char x : s.toCharArray())
            characteristic_vector[i++] = (x == c);
        return characteristic_vector;
    } // generateCharacteristicVector

    private static List<Character[]> buildParam(String[] dictionary)
    {
        List<Character[]> result = new ArrayList<Character[]>();
        for(String s : dictionary)
        {
            Character[] ca = new Character[s.length()];
            int i = 0;
            for(char c : s.toCharArray())
                ca[i++] = new Character(c);
            result.add(ca);
        }
        return result;
    } // buildParam

    /**
     * Constructs a LevenshteinAutomaton.
     * @param edit_distance maximum number of edits detected by this automaton
     * @param dictionary    list of valid input words
     */
    public LevenshteinAutomaton(int edit_distance,
            String[] dictionary)
    {
        super();
        
        this.dictionary_automaton = new DictionaryAutomaton(buildParam(dictionary));
        this.edit_distance = edit_distance;
        this.transitions = new LinkedList<Transition>();
        this.chi = DEFAULT_CHI;
        this.index_mapping = new Hashtable<State<String>,Set<Integer>>();
        this.state_mappings = new Hashtable<PositionState,State<String>>();

        Queue<PositionState> queue = new LinkedList<PositionState>();
        List<PositionState> added_states = new LinkedList<PositionState>();
//        List<boolean[]> power_set = buildPowerSet(2 * edit_distance + 2);
        List<boolean[]> power_set = buildPowerSet(2 * edit_distance + 1);

        added_states.add(START_STATE);
        state_mappings.put(START_STATE, getCurrentState());
        Set<Integer> temp_set = new TreeSet<Integer>();
        temp_set.add(new Integer(0));
        index_mapping.put(getCurrentState(), temp_set);

        // build the universal Levenshtein automaton
        queue.add(START_STATE);
        while(!queue.isEmpty())
        {
            PositionState state = queue.remove();

            for(boolean[] b : power_set)
            {
                if(coversAllPositions(edit_distance, b.length, state))
                {
                    // generate the next state
                    PositionState next_state = delta(edit_distance, state, b);

                    if(!next_state.isEmpty())
                    {
                        int index = added_states.indexOf(next_state);
                        if(index == -1)
                        {
                            // state hasn't been added before now
                            queue.add(next_state);
                            added_states.add(next_state);

                            State<String> s = new State<String>(next_state.getFirst().getParameter() == Parameter.M);
                            state_mappings.put(next_state, s);
                            temp_set = new TreeSet<Integer>();
                            for(Position pi : next_state)
                                temp_set.add(new Integer(pi.getError() - pi.getIndex()));
                            index_mapping.put(s, temp_set);
                        }
                        else
                        {
                            next_state = added_states.get(index);
                        }

                        transitions.add(new Transition(state, b, next_state));

                        State<String> first_state = state_mappings.get(state);
                        State<String> second_state = state_mappings.get(next_state);
                        if((first_state != null) && (second_state != null))
                            first_state.addNextState(convertBitStringToCharString(b), second_state);
                    }
                }
            }
        }
    } // LevenshteinAutomaton

    private static String convertBitStringToCharString(boolean[] bs)
    {
        String s = "";
        for(boolean b : bs) s += b ? "1" : "0";
        return s;
    } // convertBitStringToCharString

    private boolean[] buildCharacteristicVector(char c, String s)
    {
        int len = 2 * edit_distance + 1;//2;
        if(s.length() < len) len = s.length();
        boolean[] result = new boolean[len];
        for(int i = 0; i < len; i++)
            result[i] = (c == s.charAt(i));
        return result;
    } // buildCharacteristicVector

    private class RecognizeMapping
    {
        public String working_string;
        public Automaton<Character>.State<Character> dictionary_state;
        public Automaton<String>.State<String> levenshtein_state;
        public int index;
        public RecognizeMapping(String working_string,
                Automaton<Character>.State<Character> dictionary_state,
                Automaton<String>.State<String> levenshtein_state,
                int index)
        {
            this.working_string = working_string;
            this.dictionary_state = dictionary_state;
            this.levenshtein_state = levenshtein_state;
            this.index = index;
        }
    } // class RecognizeMapping

    private boolean isFinal(int word_length,
            int index,
            State<String> state)
    {
        for(Integer i : index_mapping.get(state))
            if(word_length - index + i.intValue() <= edit_distance)
                return true;
        return false;
    } // isFinal

    public Collection<String> recognize(String input_string)
    {
        Collection<String> result = new TreeSet<String>();
        for(int j = 0; j <= edit_distance; j++)
//      for(int j = 0; j <= 0; j++)
        {
            String s = input_string;
            for(int i = 0; i < j; i++) s = "$" + s;
            RecognizeMapping mapping = null;
            Stack<RecognizeMapping> stack = new Stack<RecognizeMapping>();
            stack.push(new RecognizeMapping("", this.dictionary_automaton.getCurrentState(), this.getCurrentState(), 0));

            while(!stack.isEmpty())
            {
                mapping = stack.pop();
                for(Character c : dictionary_automaton.getAlphabet())
                {
                    try
                    {
                        Automaton<Character>.State<Character> dictionary_next = mapping.dictionary_state.getNextState(c);
                        boolean[] characteristic_vector = buildCharacteristicVector(c.charValue(), s.substring(mapping.index));
                        Automaton<String>.State<String> levenshtein_next = mapping.levenshtein_state.getNextState(convertBitStringToCharString(characteristic_vector));

                        if(dictionary_next.isAccept() && levenshtein_next.isAccept())
                        {
                            result.add(mapping.working_string + c);
                        }

                        for(Integer k : index_mapping.get(levenshtein_next))
                        {
                            int x = k.intValue();
                            stack.push(new RecognizeMapping(mapping.working_string + c, dictionary_next, levenshtein_next, x));
                        }
                    }
                    catch(UnknownStateException us_excep){}
                }
            }
        }

        return result;
    } // recognize

    public static void main(String[] args)
    {
        String[] dictionary = { "cat", "bat", "ate", "art", "at", "matt", "kmat" };
        LevenshteinAutomaton la = new LevenshteinAutomaton(1, dictionary);
        /*
        String[] dictionary = { "xxmat", "matxx", "xmxat", "xmaxt", "xmatx", "mxatx", "maxtx",
            "alpha", "bravo", "charlie", "delta", "echo", "foxtrot", "golf" };
        LevenshteinAutomaton la = new LevenshteinAutomaton(2, dictionary);
         */
        Collection<String> r = la.recognize("mat");
    } // main
} // class LevenshteinAutomaton
